/*
        cow.h -- copy on write
        (C) Copyright 1995  John Webster Small
        All rights reserved

        Use COW<TYPE> to minimize duplicates.  For
        example, check writeReady() before modifying
        your private copy and fold() afterwards to
        reduce duplicates.

        TYPE must have:

             TYPE();
             TYPE(const TYPE&);
             stream insert/extraction operators
             int cmp(const TYPE &) const;

         If your type doesn't have a cmp() function you
         can define the following function instead:

             inline
             int COW_cmp(const TYPE& t1, const TYPE& t2)
             { return ??? }
*/

#ifndef cow_h
#define cow_h

/* LINTLIBRARY */

#ifndef cl_h
  #include "cl.h"
#endif

#ifndef single_h
  #include "single.h"
#endif

template <class TYPE>
inline int COW_cmp(const TYPE& t1, const TYPE& t2)
{ return t1.cmp(t2); }

template <class TYPE>
struct COW_OBJ {
    TYPE _t;
    unsigned links;
    COW_OBJ() : _t() {}
    COW_OBJ(const TYPE & t) : _t(t) {}
    COW_OBJ(const COW_OBJ<TYPE >& c) : _t(c._t) {}
    int cmp(const COW_OBJ<TYPE >& c) const
        { return COW_cmp(_t,c._t); }
    static int cmpTYPE2COW_OBJ
        (const TYPE * T,
         const COW_OBJ<TYPE > * C)
        { return T->cmp(C->_t); }
    CL_StrongTypingCmp(COW_OBJ<TYPE >);
};

template <class TYPE>
class COW
{
  public:
    struct Singleton_traits
        <COW<TYPE >,CL<COW_OBJ<TYPE > >,0>
    {
        static CL<COW_OBJ<TYPE > > * newInstance()
            { return new CL<COW_OBJ<TYPE > >(CL_DEL); }
    };
  private:
    Singleton<COW<TYPE >,CL<COW_OBJ<TYPE > > > instances;
    COW_OBJ<TYPE > * C;
    COW_OBJ<TYPE > * new_COW_OBJ(const TYPE& t);
  public:
    COW() : C(0) {}
    COW(const TYPE & t) : C(0) { link2(t); }
    ~COW() { unlink(); }
    void link2(const TYPE & t);
    void unlink();
    unsigned links() const
        { return (C? C->links : 0); }
    void fold();
    TYPE * writeReady();
    TYPE * readReady() const
        { return (C? &(C->_t) : 0); }
    operator TYPE *() const
        { return (C? &(C->_t) : 0); }
    operator TYPE &() const
        { assert(C); return C->_t; }
    friend ostream& operator<<
        (ostream& os, COW<TYPE >& c)
        { if (c.C) os << c.C->_t; return os; }
    friend istream& operator>>
        (istream& is, COW<TYPE >& c)
        { if (c.writeReady())
             { is >> c.C->_t; c.fold(); }
          return is; }
};

template <class TYPE>
void COW<TYPE >::link2(const TYPE & t)
{
    unlink();
    CL<COW_OBJ<TYPE > > * I = instances;
    if (I)  {
        C = I->findFirst(&t,
            CLKCCSM(COW_OBJ<TYPE >,cmpTYPE2COW_OBJ));
        if (!C)
            C = I->insSortDL(new COW_OBJ<TYPE >(t));
        if (C)
            C->links++;
    }
}

template <class TYPE>
void COW<TYPE >::unlink()
{
    if (C)  {
        assert(C->links);
        if (!--C->links)  {
            CL<COW_OBJ<TYPE > > * I = instances.ok();
            if (I && I->findFirst(C))
                do {
                    if (C == I->get())  {
                        I->delBackup();
                        break;
                    }
                } while (I->findNext(C));
        }
        C = 0;
    }
}

template <class TYPE>
void COW<TYPE >::fold()
{
    if (C && C->links == 1)  {
        CL<COW_OBJ<TYPE > > * I = instances.ok();
        assert(I);
        if (I->findFirst(C))
            if (C == I->get())  {
                unsigned discard = I->curNode();
                if (I->findNext(C))  {
                    C = I->get();
                    C->links++;
                    I->atDel(discard);
                }
            }
            else  {
                unsigned save = I->curNode();
                while (I->findNext(C))
                    if (C == I->get())  {
                        C = I->atGet(save);
                        C->links++;
                        I->delBackup();
                        break;
                    }
            }
    }
}

template <class TYPE>
TYPE * COW<TYPE >::writeReady()
{
    CL<COW_OBJ<TYPE > > * I = instances;
    if (I)  {
        if (!C)  {
            if ((C = I->insSortDL(new COW_OBJ<TYPE >))
                == 0)
                return 0;
        }
        else if (C->links > 1)  {
            COW_OBJ<TYPE > * _C =
                I->insSortDL(new COW_OBJ<TYPE >(C->_t));
            if (!_C)
                return 0;
            C->links--;
            C = _C;
        }
        C->links = 1;
        return &(C->_t);
    }
    return 0;
}

#endif
